import sofia.micro.*;
import sofia.graphics.Image;
import sofia.util.Random;
import java.util.Set;

//-------------------------------------------------------------------------
/**
 * Wombat. A Wombat moves forward until it can't do so anymore, at
 * which point it turns left. This wombat can not move over rocks.
 * If a wombat finds a leaf, it eats it.
 *
 * @author Michael Kolling
 * @author  Stephen Edwards
 * @author  Last changed by $Author: edwards $
 * @version $Date: 2012/08/23 15:35 $
 */
public class Wombat
    extends Actor
{
    //~ Fields ................................................................

    private static final int EAST = 0;
    private static final int WEST = 1;
    private static final int NORTH = 2;
    private static final int SOUTH = 3;

    private int direction;
    private int leavesEaten;

    private Image wombatRight;
    private Image wombatLeft;


    //~ Constructor ...........................................................

    // ----------------------------------------------------------
    /**
     * Create a new leaf.
     */
    public Wombat()
    {
        wombatRight = getImage();
        wombatLeft = new Image("wombat-left.gif");

        setDirection(EAST);
        leavesEaten = 0;
    }


    // ----------------------------------------------------------
    /**
     * Do whatever the wombat likes to to just now.
     */
    public void act()
    {
        if (foundLeaf())
        {
            eatLeaf();
        }
        else if (canMove())
        {
            move();
        }
        else
        {
            turnRandom();
        }
    }


    // ----------------------------------------------------------
    /**
     * Check whether there is a leaf in the same cell as we are.
     * @return True if there is a leaf in our current cell.
     */
    public boolean foundLeaf()
    {
        Actor leaf = getOneObjectAtOffset(0, 0, Leaf.class);
        if (leaf != null)
        {
            return true;
        }
        else
        {
            return false;
        }
    }


    // ----------------------------------------------------------
    /**
     * Eat a leaf.
     */
    public void eatLeaf()
    {
        Actor leaf = getOneObjectAtOffset(0, 0, Leaf.class);
        if (leaf != null)
        {
            // eat the leaf...
            getWorld().remove(leaf);
            leavesEaten = leavesEaten + 1;
        }
    }


    // ----------------------------------------------------------
    /**
     * Move one cell forward in the current direction.
     */
    public void move()
    {
        if (!canMove())
        {
            return;
        }
        switch (direction)
        {
            case SOUTH :
                setGridY(getGridY() + 1);
                break;
            case EAST :
                setGridX(getGridX() + 1);
                break;
            case NORTH :
                setGridY(getGridY() - 1);
                break;
            case WEST :
                setGridX(getGridX() - 1);
                break;
        }
    }


    // ----------------------------------------------------------
    /**
     * Test if we can move forward. Return true if we can, false otherwise.
     * @return True if we can move forward.
     */
    public boolean canMove()
    {
        World myWorld = getWorld();
        int x = getGridX();
        int y = getGridY();
        switch (direction)
        {
            case SOUTH :
                y++;
                break;
            case EAST :
                x++;
                break;
            case NORTH :
                y--;
                break;
            case WEST :
                x--;
                break;
        }
        // test for outside border
        if (x >= myWorld.getWidth() || y >= myWorld.getHeight())
        {
            return false;
        }
        else if (x < 0 || y < 0)
        {
            return false;
        }
        Set<Rock> rocks = myWorld.getObjectsAt(x, y, Rock.class);
        if (rocks.isEmpty())
        {
            return true;
        }
        else
        {
            return false;
        }
    }


    // ----------------------------------------------------------
    /**
     * Turn in a random direction.
     */
    public void turnRandom()
    {
        // get a random number between 0 and 3...
        int turns = Random.generator().nextInt(4);

        // ...an turn left that many times.
        for (int i = 0; i < turns; i++)
        {
            turnLeft();
        }
    }


    // ----------------------------------------------------------
    /**
     * Turns towards the left.
     */
    public void turnLeft()
    {
        switch (direction)
        {
            case SOUTH :
                setDirection(EAST);
                break;
            case EAST :
                setDirection(NORTH);
                break;
            case NORTH :
                setDirection(WEST);
                break;
            case WEST :
                setDirection(SOUTH);
                break;
        }
    }


    // ----------------------------------------------------------
    /**
     * Sets the direction we're facing.
     * @param direction The new direction to face.
     */
    public void setDirection(int direction)
    {
        this.direction = direction;
        switch (direction)
        {
            case SOUTH :
                setImage(wombatRight);
                setRotation(90);
                break;
            case EAST :
                setImage(wombatRight);
                setRotation(0);
                break;
            case NORTH :
                setImage(wombatLeft);
                setRotation(90);
                break;
            case WEST :
                setImage(wombatLeft);
                setRotation(0);
                break;
            default :
                break;
        }
    }


    // ----------------------------------------------------------
    /**
     * Tell how many leaves we have eaten.
     * @return The number of leaves eaten so far.
     */
    public int getLeavesEaten()
    {
        return leavesEaten;
    }
}
